package core.engine.toolbox;

import core.engine.Engine;
import core.engine.entity.Camera;
import core.engine.terrain.Terrain;
import core.engine.terrain.TerrainTile;
import org.lwjgl.system.windows.MOUSEINPUT;
import utils.MatrixGraphicUtils;
import utils.vectors.*;

public class MousePicker {

    private static final int RECURSION_COUNT = 200;
    private static final float RAY_RANGE = 600;

    private Vector3f currentRay = new Vector3f();

    private Matrix4f projectionMatrix;
    private Matrix4f viewMatrix;
    private Camera camera;

    private Terrain terrain;
    private Vector3f currentTerrainPoint;
    private TerrainTile currentTerrainTile;

    public MousePicker(Camera cam, Matrix4f projection, Terrain terrain) {
        camera = cam;
        projectionMatrix = projection;
        viewMatrix = MatrixGraphicUtils.createViewMatrix(camera);
        this.terrain = terrain;
    }

    public Vector3f getCurrentTerrainPoint() {
        return currentTerrainPoint;
    }

    public Vector3f getCurrentRay() {
        return currentRay;
    }

    public void update(float mouseX, float mouseY) {
        viewMatrix = MatrixGraphicUtils.createViewMatrix(camera);
        currentRay = calculateMouseRay(mouseX, mouseY);
        currentTerrainPoint = calcCurrentTerrainIntersection();
    }

    private Vector3f calculateMouseRay(float mouseX, float mouseY) {

        Vector2f normalizedCoords = getNormalisedDeviceCoordinates(mouseX, mouseY);
        Vector4f clipCoords = new Vector4f(normalizedCoords.x, normalizedCoords.y, -1.0f, 1.0f);
        Vector4f eyeCoords = toEyeCoords(clipCoords);
        return toWorldCoords(eyeCoords);
    }

    private Vector3f toWorldCoords(Vector4f eyeCoords) {
        Matrix4f invertedView = Matrix4f.invert(viewMatrix, null);
        Vector4f rayWorld = Matrix4f.transform(invertedView, eyeCoords, null);
        Vector3f mouseRay = new Vector3f(rayWorld.x, rayWorld.y, rayWorld.z);
        mouseRay.normalise();
        return mouseRay;
    }

    private Vector4f toEyeCoords(Vector4f clipCoords) {
        Matrix4f invertedProjection = Matrix4f.invert(projectionMatrix, null);
        Vector4f eyeCoords = Matrix4f.transform(invertedProjection, clipCoords, null);
        return new Vector4f(eyeCoords.x, eyeCoords.y, -1f, 0f);
    }

    private Vector2f getNormalisedDeviceCoordinates(float mouseX, float mouseY) {
        float x = (2.0f * mouseX) / Engine.WINDOW_WIDTH - 1f;
        float y = (2.0f * mouseY) / Engine.WINDOW_HEIGHT - 1f;
        return new Vector2f(x, y);
    }

    //**********************************************************

    private Vector3f calcCurrentTerrainIntersection() {
        Vector3f camPos = camera.getPosition();
        return null;
    }

    //**********************************************************


}
